"""LUos.py"""
# -*- coding: UTF-8 -*-
__annotations__ = """
 =======================================================
 Copyright (c) 2023-2025
 Author:
     Lisitsin Y.R.
 Project:
     LU_PY
     Python (LU)
 Module:
     LUos.py

 =======================================================
"""

#------------------------------------------
# БИБЛИОТЕКИ python
#------------------------------------------
import os
import platform
import sys
import enum
import ctypes
import datetime
import subprocess

"""
Кроссплатформенные функции:
Информации об архитектуре platform.architecture(),
Тип машины platform.machine(),
Сетевое имя компьютера platform.node(),
Сведения о базовой платформе platform.platform(),
Реальное имя процессора platform.processor(),
Номер и дата сборки Python platform.python_build(),
Версия компилятора platform.python_compiler(),
Ветвь SCM реализации Python platform.python_branch(),
Реализация Python platform.python_implementation(),
Ревизия SCM реализации Python platform.python_revision(),
Версия Python как строка platform.python_version(),
Версия Python как кортеж platform.python_version_tuple(),
Сведения о выпуске системы platform.release(),
Имя операционной системы platform.system(),
platform.system_alias(),
Версия выпуска системы platform.version(),
Сведения команды терминала uname platform.uname(),
Функции платформы Java:
Версия интерфейса для Jython platform.java_ver(),
Функции платформы Windows:
Информация о версии из реестра Windows platform.win32_ver(),
Текущая редакция Windows platform.win32_edition(),
True, если Windows, распознается как IoT platform.win32_is_iot(),
Функции платформы Mac OS:
Информация о версии Mac OS platform.mac_ver(),
Функции платформы Unix:
Версия библиотеки libc platform.libc_ver(),
Функции платформы Linux:
Идентификатор ОС из os-release platform.freedesktop_os_release() (доступна в Python 3.10).
"""

#------------------------------------------
# БИБЛИОТЕКИ сторонние
#------------------------------------------
import psutil

# if platform.system() == 'Windows':
#     import win32api
#     import win32con
# #endif

#------------------------------------------
# БИБЛИОТЕКИ LU
#------------------------------------------
if platform.system() == 'Windows':
    import lyrpy.LUParserREG as LUParserREG
#endif

import lyrpy.LUDateTime as LUDateTime

#Correspondence to tools in the os module
#-----------------------------------------------------------------------------
# os and os.path                      pathlib
# -----------------------------------------------------------------------------
# os.path.abspath()                   Path.absolute() [1]
# os.path.realpath()                  Path.resolve()
# os.chmod()                          Path.chmod()
# os.mkdir()                          Path.mkdir()
# os.makedirs()                       Path.mkdir()
# os.rename()                         Path.rename()
# os.replace()                        Path.replace()
# os.rmdir()                          Path.rmdir()
# os.remove(), os.unlink()            Path.unlink()
# os.getcwd()                         Path.cwd()
# os.path.exists()                    Path.exists()
# os.path.expanduser()                Path.expanduser() and Path.home()
# os.listdir()                        Path.iterdir()
# os.walk()                           Path.walk()
# os.path.isdir()                     Path.is_dir()
# os.path.isfile()                    Path.is_file()
# os.path.islink()                    Path.is_symlink()
# os.link()                           Path.hardlink_to()
# os.symlink()                        Path.symlink_to()
# os.readlink()                       Path.readlink()
# os.path.relpath()                   PurePath.relative_to() [2]
# os.stat()                           Path.stat(), Path.owner(), Path.group()
# os.path.isabs()                     PurePath.is_absolute()
# os.path.join()                      PurePath.joinpath()
# os.path.basename()                  PurePath.name
# os.path.dirname()                   PurePath.parent
# os.path.samefile()                  Path.samefile()
# os.path.splitext()                  PurePath.stem and PurePath.suffix


cHOME = 'HOME'
cWINDIR = 'windir'
cTEST = 'TEST'

@enum.unique
class TProductType(enum.Enum):
    """TFoldersConst"""
    W95 = 'Windows 95'
    W98 = 'Windows 98'
    WNT1 = 'Windows NT Workstation'
    WNT2 = 'Windows NT Server'
    WNT3 = 'Windows NT Domain Controller'
    W20001 = 'Windows 2000 Professional'
    W20002 = 'Windows 2000 Server'
    W20003 = 'Windows 2000 Domain Controller'
    WXP1 = 'Windows XP Home Edition'
    WXP2 = 'Windows XP Home Edition Tablet PC'
    WXP3 = 'Windows XP Professional'
    WXP4 = 'Windows XP Professional Tablet PC'
    W20031 = 'Windows Server 2003'
    W20032 = 'Windows Server 2003 Domain Controller'
    W7 = 'Windows 6.1 / 1'
    W8 = 'Windows 6.2 / 1'
    WXX = 'W95 or W98'
    WXXXX = 'W2000 or WXP or W2003'
#endclass

cSection = 'SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\Shell Folders'

class TFoldersConst(enum.Enum):
    """TFoldersConst"""
    cuDesktop     = 'Desktop'
    cuStartMenu   = 'Start Menu'
    cuFavorites   = 'Favorites'
    cuPrograms    = 'Programs'
    cuStartup     = 'Startup' 
    cuMyDocuments = 'Personal'

    lmDesktop     = 'Desktop'
    lmStartMenu   = 'Start Menu'
    lmFavorites   = 'Favorites'
    lmPrograms    = 'Programs'
    lmStartup     = 'Startup' 
#endclass

class TFolders (object):
    """TFolders"""
    luClassName = 'TFolders'
    __annotations__ ="""
    """
    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, **kwargs):
        """ Constructor """
    #beginfunction
        super ().__init__ (**kwargs)
        self.__FRegParser = LUParserREG.TREGParser()
        ...
    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__ (self):
        """ destructor """
    #beginfunction
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format(LClassName)
        # LULog.LoggerTOOLS_AddLevel (LULog.DEBUGTEXT, s)
        #print (s)
    #endfunction

    #--------------------------------------------------
    # @property cuDesktop
    #--------------------------------------------------
    # getter
    @property
    def cuDesktop(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuDesktop.value)
    #endfunction
    # setter
    @cuDesktop.setter
    def cuDesktop (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuDesktop.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property lmDesktop
    #--------------------------------------------------
    # getter
    @property
    def lmDesktop(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuDesktop.value)
    #endfunction
    # setter
    @lmDesktop.setter
    def lmDesktop (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuDesktop.value, AValue)
    #endfunction
    #--------------------------------------------------
    # @property cuStartMenu
    #--------------------------------------------------
    # getter
    @property
    def cuStartMenu(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuStartMenu.value)
    #endfunction
    # setter
    @cuStartMenu.setter
    def cuStartMenu (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuStartMenu.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property lmStartMenu
    #--------------------------------------------------
    # getter
    @property
    def lmStartMenu(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuStartMenu.value)
    #endfunction
    # setter
    @lmStartMenu.setter
    def lmStartMenu (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuStartMenu.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property cuFavorites
    #--------------------------------------------------
    # getter
    @property
    def cuFavorites(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuFavorites.value)
    #endfunction
    # setter
    @cuFavorites.setter
    def cuFavorites (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuFavorites.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property lmFavorites
    #--------------------------------------------------
    # getter
    @property
    def lmFavorites(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuFavorites.value)
    #endfunction
    # setter
    @lmFavorites.setter
    def lmFavorites (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuFavorites.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property cuPrograms
    #--------------------------------------------------
    # getter
    @property
    def cuPrograms(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuPrograms.value)
    #endfunction
    # setter
    @cuPrograms.setter
    def cuPrograms (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuPrograms.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property lmPrograms
    #--------------------------------------------------
    # getter
    @property
    def lmPrograms(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuPrograms.value)
    #endfunction
    # setter
    @lmPrograms.setter
    def lmPrograms (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuPrograms.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property cuStartup
    #--------------------------------------------------
    # getter
    @property
    def cuStartup(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuStartup.value)
    #endfunction
    # setter
    @cuStartup.setter
    def cuStartup (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuStartup.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property lmStartup
    #--------------------------------------------------
    # getter
    @property
    def lmStartup(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuStartup.value)
    #endfunction
    # setter
    @lmStartup.setter
    def lmStartup (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuStartup.value, AValue)
    #endfunction

    #--------------------------------------------------
    # @property cuMyDocuments
    #--------------------------------------------------
    # getter
    @property
    def cuMyDocuments(self):
    #beginfunction
        return self.GetFolderCU (TFoldersConst.cuMyDocuments.value)
    #endfunction
    # setter
    @cuMyDocuments.setter
    def cuMyDocuments (self, AValue: str):
    #beginfunction
        self.SetFolderCU (TFoldersConst.cuMyDocuments.value, AValue)
    #endfunction

    #--------------------------------------------------------------------------------
    # GetFolderCU (AFolderName: str) -> str:
    #--------------------------------------------------------------------------------
    def GetFolderCU (self, AFolderName: str) -> str:
    #beginfunction
        LKeyName = AFolderName
        LFolderValue, LType = self.__FRegParser.GetKeyReg (LUParserREG.THKEYConst.cHKCU, cSection, LKeyName)
        return LFolderValue
    #endfunction

    #--------------------------------------------------------------------------------
    # SetFolderCU (AFolderName: str, AValue: str)
    #--------------------------------------------------------------------------------
    def SetFolderCU (self, AFolderName: str, AValue: str):
    #beginfunction
        LKeyName = AFolderName
        self.__FRegParser.SetValueReg (LUParserREG.THKEYConst.cHKCU, cSection,
                                       LKeyName, LUParserREG.TValueTypes.vtEXPAND_SZ, AValue)
    #endfunction

    #--------------------------------------------------------------------------------
    # GetFolderLM (AFolderName: str) -> str:
    #--------------------------------------------------------------------------------
    def GetFolderLM (self, AFolderName: str) -> str:
    #beginfunction
        LKeyName = AFolderName
        LFolderValue, LType = self.__FRegParser.GetKeyReg(LUParserREG.THKEYConst.cHKLM, cSection, LKeyName)
        return LFolderValue
    #endfunction

    #--------------------------------------------------------------------------------
    # SetFolderLM (AFolderName: str, AValue: str)
    #--------------------------------------------------------------------------------
    def SetFolderLM (self, AFolderName: str, AValue: str):
    #beginfunction
        LKeyName = AFolderName
        self.__FRegParser.SetValueReg (LUParserREG.THKEYConst.cHKLM, cSection,
                                LKeyName, LUParserREG.TValueTypes.vtEXPAND_SZ, AValue)
    #endfunction
#endclass

class TOSInfo (object):
    """TOSInfo"""
    luClassName = 'TOSInfo'
    __annotations__ ="""
    """
    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, **kwargs):
        """ Constructor """
    #beginfunction
        super ().__init__ (**kwargs)
        ...
    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__ (self):
        """ destructor """
    #beginfunction
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format(LClassName)
        #LUConst.LULogger.log (LULog.DEBUGTEXT, s)
    #endfunction

    # Return the number of CPUs in the system. Returns None if undetermined.
    #--------------------------------------------------
    # @property CPUs
    #--------------------------------------------------
    # getter
    @property
    def CPUs (self):
    #beginfunction
        return os.cpu_count()
    #endfunction

    # Returns the machine type
    #--------------------------------------------------
    # @property machine
    #--------------------------------------------------
    # getter
    @property
    def machine (self):
    #beginfunction
        return platform.machine()
    #endfunction

    # Returns the computer’s network name
    #--------------------------------------------------
    # @property node
    #--------------------------------------------------
    # getter
    @property
    def node (self):
    #beginfunction
        return platform.node()
    #endfunction

    # Returns a namedtuple() containing six attributes: system, node, release, version, machine, and processor
    #--------------------------------------------------
    # @property uname
    #--------------------------------------------------
    # getter
    @property
    def uname (self):
    #beginfunction
        return platform.uname ()
    #endfunction

    # Returns the (real) processor name, e.g. 'amdk6'.
    #--------------------------------------------------
    # @property processor
    #--------------------------------------------------
    # getter
    @property
    def processor (self):
    #beginfunction
        return platform.processor ()
    #endfunction

    # Returns the system’s release
    #--------------------------------------------------
    # @property release
    #--------------------------------------------------
    # getter
    @property
    def release (self):
    #beginfunction
        return platform.release ()
    #endfunction

    # Returns the system/OS name, such as 'Linux', 'Darwin', 'Java', 'Windows'
    #--------------------------------------------------
    # @property release
    #--------------------------------------------------
    # getter
    @property
    def system (self):
    #beginfunction
        return platform.system ()
    #endfunction

    # Returns the system’s release version
    #--------------------------------------------------
    # @property version
    #--------------------------------------------------
    # getter
    @property
    def version (self):
    #beginfunction
        return platform.version ()
    #endfunction

    #Returns (system, release, version) aliased to common marketing names
    #platform.system_alias(system, release, version)

    # Get additional version information from the Windows Registry and return a tuple (release, version, csd, ptype) referring to OS release
    #--------------------------------------------------
    # @property win32_ver
    #--------------------------------------------------
    # getter
    @property
    def win32_ver (self):
    #beginfunction
        return platform.win32_ver()
    #endfunction

    # Returns a string representing the current Windows edition, or None if the value cannot be determined. Possible values include but are not limited to 'Enterprise', 'IoTUAP', 'ServerStandard', and 'nanoserver'.
    #--------------------------------------------------
    # @property win32_edition
    #--------------------------------------------------
    @property
    def win32_edition (self):
    #beginfunction
        return platform.win32_edition ()
    #endfunction

    #---------------------------------------------
    # Python
    #---------------------------------------------
    # Returns a tuple (buildno, builddate) stating the Python build number and date as strings.
    #--------------------------------------------------
    # @property python_build
    #--------------------------------------------------
    @property
    def python_build (self):
    #beginfunction
        return platform.python_build ()
    #endfunction

    # Returns a string identifying the compiler used for compiling Python.
    #--------------------------------------------------
    # @property python_compiler
    #--------------------------------------------------
    @property
    def python_compiler (self):
    #beginfunction
        return platform.python_compiler ()
    #endfunction

    # Returns a string identifying the Python implementation SCM branch.
    #--------------------------------------------------
    # @property python_branch
    #--------------------------------------------------
    @property
    def python_branch (self):
    #beginfunction
        return platform.python_branch ()
    #endfunction

    # Returns a string identifying the Python implementation
    #--------------------------------------------------
    # @property python_implementation
    #--------------------------------------------------
    @property
    def python_implementation (self):
    #beginfunction
        return platform.python_implementation ()
    #endfunction

    # Returns a string identifying the Python implementation SCM revision.
    #--------------------------------------------------
    # @property python_revision
    #--------------------------------------------------
    @property
    def python_revision (self):
    #beginfunction
        return platform.python_revision ()
    #endfunction

    # Returns the Python version
    #--------------------------------------------------
    # @property python_version
    #--------------------------------------------------
    @property
    def python_version (self):
    #beginfunction
        return platform.python_version ()
    #endfunction

    # Python sys.version
    #--------------------------------------------------
    # @property version_sys
    #--------------------------------------------------
    @property
    def version_sys (self):
    #beginfunction
        return sys.version
    #endfunction

    # Returns the Python version as tuple (major, minor, patchlevel) of strings.
    #--------------------------------------------------
    # @property python_version_tuple
    #--------------------------------------------------
    @property
    def python_version_tuple (self):
    #beginfunction
        return platform.python_version_tuple ()
    #endfunction

    #--------------------------------------------------
    # @property UserID
    #--------------------------------------------------
    @property
    def UserID (self):
    #beginfunction
        return os.getlogin ()
    #endfunction

    @staticmethod
    def Info_psutil ():
        """Info_psutil"""
    #beginfunction
        """CPU"""
        # Return system CPU times as a named tuple
        psutil.cpu_times(percpu=False)
        # Return the number of logical CPUs in the system
        psutil.cpu_count (logical = True)

        """Memory"""
        # Return statistics about system memory usage as a named tuple including the following fields, expressed in bytes
        psutil.virtual_memory ()

        """Disks"""
        # Return all mounted disk partitions as a list of named tuples including device, mount point and filesystem type
        psutil.disk_partitions (all = False)
        # Return disk usage statistics about the partition which contains the given path as a named tuple including total, used and free space
        # psutil.disk_usage (path)

        """Network"""
        # Return the addresses associated to each NIC (network interface card) installed on the system as a dictionary whose keys are the NIC names and value is a list of named tuples for each address assigned to the NIC
        psutil.net_if_addrs()

        """Other system info"""
        psutil.users ()

        """Processes"""

        """Windows services"""
    #endfunction
#endclass

#------------------------------------------
# GetCurrentDir
#------------------------------------------
def GetCurrentDir () -> str:
    """GetCurrentDir"""
#beginfunction
    return os.getcwd ()
#endfunction

#------------------------------------------
# APPWorkDir
#------------------------------------------
def APPWorkDir () -> str:
    """APPWorkDir"""
#beginfunction
    return os.getcwd ()
#endfunction

#------------------------------------------
# GetEnvVar
#------------------------------------------
def GetEnvVar (AEnvVar: str) -> str:
    """GetEnvVar"""
#beginfunction
    s = ''
    try:
        s = os.environ [AEnvVar]
    except:
        ...
    finally:
        ...
    return s
#endfunction

#------------------------------------------
# SetEnvVar
#------------------------------------------
def SetEnvVar (AEnvVar: str, AValue: str):
    """SetEnvVar"""
#beginfunction
    os.environ[AEnvVar] = AValue
#endfunction

#------------------------------------------
# get_data
#------------------------------------------
def get_data (EXTENDED_NAME_FORMAT: int):
#beginfunction
    GetUserNameEx = ctypes.windll.secur32.GetUserNameExW
    data = EXTENDED_NAME_FORMAT
    size = ctypes.pointer (ctypes.c_ulong (0))
    GetUserNameEx (data, None, size)
    nameBuffer = ctypes.create_unicode_buffer (size.contents.value)
    GetUserNameEx (data, nameBuffer, size)
    return nameBuffer.value
#endfunction

#------------------------------------------
# get_display_name
#------------------------------------------
def get_display_name ():
#beginfunction
    GetUserNameEx = ctypes.windll.secur32.GetUserNameExW
    NameDisplay = 3
    size = ctypes.pointer (ctypes.c_ulong (0))
    GetUserNameEx (NameDisplay, None, size)
    nameBuffer = ctypes.create_unicode_buffer (size.contents.value)
    GetUserNameEx (NameDisplay, nameBuffer, size)
    return nameBuffer.value
#endfunction

#------------------------------------------
# Print_get_data
#------------------------------------------
def Print_get_data ():
#beginfunction
    print ("NameUnknown            : ", get_data (0))
    print ("NameFullyQualifiedDN   : ", get_data (1))
    print ("NameSamCompatible      : ", get_data (2))
    print ("NameDisplay            : ", get_data (3))
    print ("NameUniqueId           : ", get_data (6))
    print ("NameCanonical          : ", get_data (7))
    print ("NameUserPrincipal      : ", get_data (8))
    print ("NameCanonicalEx        : ", get_data (9))
    print ("NameServicePrincipal   : ", get_data (10))
    print ("NameDnsDomain          : ", get_data (12))
#endfunction

#------------------------------------------------------
# PrintGeneralTitle
#------------------------------------------------------
def PrintGeneralTitle ():
#beginfunction
    LTOSInfo = TOSInfo()
    print ('===========================================')
    print ('Текущее время = ' + LUDateTime.DateTimeStr (True, datetime.datetime.now(), LUDateTime.cFormatDateTimeLog05))
    print ('UserID        = ' + LTOSInfo.UserID)
    # print ('FullName      = ' + @FullName+' ('+@Comment+')')
    print ('PCUser        = ' + LTOSInfo.node)
    print ('CPUs          = ' + str(LTOSInfo.CPUs))
    # print (LTOSInfo.uname)
    # print ('CPU           = ' + @CPU+' '+@MHz)
    print ('HostName      = ' + LTOSInfo.node)
    print ('OS            = ' + LTOSInfo.uname.system+' '+ '('+LTOSInfo.uname.version+')')
    # print ('CompName      = ' + @WKSTA)
    # print ('===========================================')
    # print ('DomainPC      = ' + @Domain)
    # print ('DomainUser    = ' + @LDomain)
    # print ('LServer       = ' + @LServer)
    print ('===========================================')
#endfunction

#------------------------------------------
# Exec_dir ():
#------------------------------------------
def Exec_dir ():
#beginfunction
    # Пример: Выполнение команды dir и получение вывода
    result = subprocess.run(["dir"], shell=True, capture_output=True, text=True)
    print(result.stdout)  # Вывод команды
#endfunction

#------------------------------------------------------
# main
#------------------------------------------------------
def main ():
#beginfunction
    # PrintGeneralTitle ()
    # Print_get_data ()
    # get_display_name ()
    ...
#endfunction

GOSInfo = TOSInfo ()

#------------------------------------------
#
#------------------------------------------
#beginmodule
if __name__ == '__main__':
    main()
#endif

#endmodule



"""LUSheduler.py"""
# -*- coding: UTF-8 -*-
__annotations__ = """
 =======================================================
 Copyright (c) 2023-2024
 Author:
     Lisitsin Y.R.
 Project:
     LU_PY
     Python (LU)
 Module:
     LUSheduler.py

 =======================================================
"""

#------------------------------------------
# БИБЛИОТЕКИ python
#------------------------------------------
import datetime
from calendar import monthrange
import threading
import logging

#------------------------------------------
# БИБЛИОТЕКИ сторонние
#------------------------------------------

#------------------------------------------
# БИБЛИОТЕКИ LU
#------------------------------------------
import lyrpy.LULog as LULog
import lyrpy.LUStrUtils as LUStrUtils
import lyrpy.LUDateTime as LUDateTime

minNN = 0
maxNN = 59
minHH = 0
maxHH = 23
minDW = 1
maxDW = 7
minMM = 1
maxMM = 12
minDD = 1
maxDD = 31
DelimPattern = ' '
DelimItem = ','
DelimItems = '-'
"""
------------------------------------------------------------------
 NN-NN    HH-HH     DD-DD     MM-MM     DW-DW     [Program]
 NN,NN,.. HH,HH,... DD,DD,... MM,MM,... DW,DW,... [Program]
 *        *         *         *         *         [Program]
------------------------------------------------------------------
"""

def NotifyFileEvent():
#beginfunction
    ...
#endfunction
TNotifyFileEvent = NotifyFileEvent

# --------------------------------------------
# TObjectsItem
# --------------------------------------------
class TShedulerEventItem (object):
    """TShedulerEventItem"""
    luClassName = 'TShedulerEventItem'

    @staticmethod
    def IsBitOn (Value: int, Bit: int) -> bool:
    #beginfunction
        # return n & (1<<b))
        LResult = (Value & (1<<Bit)) != 0
        return LResult
    #endfunction

    @staticmethod
    def TurnBitOn (Value: int, Bit: int) -> int:
    #beginfunction
        #return value | (1<<bit)
        LResult = Value | (1<<Bit)
        return LResult
    #endfunction

    @staticmethod
    def TurnBitOff (Value: int, Bit: int) -> int:
    #beginfunction
        # return value & ~(1<<bit)
        LResult = Value & ~ (1<<Bit)
        return LResult
    #endfunction

    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__(self):
        """ Constructor """
        super().__init__()
        self.__FDD: int = 0
        self.__FMM: int = 0
        self.__FDW: int = 0
        self.__FHH: int = 0
        self.__FNN1: int = 0
        self.__FNN2: int = 0
        self.__FList: () = list ()
        self.__FNameEvent: str = ''
        self.Clear()

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__(self):
        """ destructor """
    #beginfunction
        # удалить объект
        del self.__FList
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format (LClassName)
        # LUConst.LULogger.log (LULog.DEBUGTEXT, s)
        #print (s)
    #endfunction

    def Clear(self):
    #beginfunction
        ...
    #endfunction

    #--------------------------------------------------
    # @property NameEvent
    #--------------------------------------------------
    # getter
    @property
    def NameEvent(self):
    #beginfunction
        return self.__FNameEvent
    #endfunction
    @NameEvent.setter
    def NameEvent(self, Value: str):
    #beginfunction
        self.__FNameEvent: str = Value
    #endfunction

    def GetXX (self) -> ():
    #beginfunction
        return self.__FNN1,self.__FNN2,self.__FHH,self.__FDD,self.__FMM,self.__FDW
    #endfunction

    #--------------------------------------------------
    # @property NN
    #--------------------------------------------------
    # property NN [Index: Integer]: Boolean read GetNN write SetNN;
    def GetNN (self, Index: int) -> bool:
    #beginfunction
        LResult = False
        if (Index >= minNN) and (Index <= maxNN):
            if Index <= 31:
                LResult = self.IsBitOn (self.__FNN1, Index)
            else:
                LResult = self.IsBitOn (self.__FNN2, Index-32)
            #endif
        #endif
        return LResult
    #endfunction
    def SetNN (self, Index: int, Value: bool):
    #beginfunction
        if (Index >= minNN) and (Index <= maxNN):
            if Index <= 31:
                if Value:
                    self.__FNN1 = self.TurnBitOn (self.__FNN1, Index)
                else:
                    self.__FNN1 = self.TurnBitOff (self.__FNN1, Index)
                #endif
            else:
                if Value:
                    self.__FNN2 = self.TurnBitOn (self.__FNN2, Index-32)
                else:
                    self.__FNN2 = self.TurnBitOff (self.__FNN2, Index-32)
                #endif
            #endif
        #endif
    #endfunction

    #--------------------------------------------------
    # @property HH
    #--------------------------------------------------
    # property HH [Index: Integer]: Boolean read GetHH write SetHH;
    def GetHH (self, Index: int) -> bool:
    #beginfunction
        LResult = False
        if (Index >= minHH) and (Index <= maxHH):
            LResult = self.IsBitOn (self.__FHH, Index)
        return LResult
    #endfunction
    def SetHH (self, Index: int, Value: bool):
    #beginfunction
        if (Index >= minHH) and (Index <= maxHH):
            if Value:
                self.__FHH = self.TurnBitOn (self.__FHH, Index)
            else:
                self.__FHH = self.TurnBitOff (self.__FHH, Index)
            #endif
        #endif
    #endfunction

    #--------------------------------------------------
    # @property DD
    #--------------------------------------------------
    # property DD [Index: Integer]: Boolean read GetDD write SetDD;
    def GetDD (self, Index: int) -> bool:
    #beginfunction
        LResult = False
        if (Index >= minDD-1) and (Index <= maxDD-1):
        # if (Index >= minDD - 1) and (Index <= maxDD):
           LResult = self.IsBitOn (self.__FDD, Index)
        return LResult
    #endfunction
    def SetDD (self, Index: int, Value: bool):
    #beginfunction
        if (Index >= minDD-1) and (Index <= maxDD-1):
            if Value:
                self.__FDD = self.TurnBitOn (self.__FDD, Index)
            else:
                self.__FDD = self.TurnBitOff (self.__FDD, Index)
            #endif
        #endif
    #endfunction

    #--------------------------------------------------
    # @property MM
    #--------------------------------------------------
    # property MM [Index: Integer]: Boolean read GetMM write SetMM;
    def GetMM (self, Index: int) -> bool:
    #beginfunction
        LResult = False
        if (Index >= minMM-1) and (Index <= maxMM-1):
            LResult = self.IsBitOn (self.__FMM, Index)
        return LResult
    #endfunction
    def SetMM (self, Index: int, Value: bool):
    #beginfunction
        if (Index >= minMM-1) and (Index <= maxMM-1):
            if Value:
                self.__FMM = self.TurnBitOn (self.__FMM, Index)
            else:
                self.__FMM = self.TurnBitOff (self.__FMM, Index)
            #endif
        #endif
    #endfunction

    #--------------------------------------------------
    # @property DW
    #--------------------------------------------------
    # property DW [Index: Integer]: Boolean read GetDW write SetDW;
    def GetDW (self, Index: int) -> bool:
    #beginfunction
        LResult = False
        if (Index >= minDW-1) and (Index <= maxDW-1):
            LResult = self.IsBitOn (self.__FDW, Index)
        return LResult
    #endfunction
    def SetDW (self, Index: int, Value: bool):
    #beginfunction
        if (Index >= minDW-1) and (Index <= maxDW-1):
            if Value:
                self.__FDW = self.TurnBitOn (self.__FDW, Index)
            else:
                self.__FDW = self.TurnBitOff (self.__FDW, Index)
            #endif
        #endif
    #endfunction

    #--------------------------------------------------
    # @property XXString
    #--------------------------------------------------
    # property XXString[XXName: string]: string read GetXXString;
    def GetXXString (self, XXName: str) -> str:
    #beginfunction
        LResult = ''
        if XXName.upper() == 'DD':
            for i in range (0, maxDD):
                if self.GetDD(i):
                    LResult = LResult+LUStrUtils.AddChar(' ', str(i+1), 2)+','
                else:
                    LResult = LResult + 'xx' + ','
                #endif
            #endfor
        #endif
        if XXName.upper() == 'MM':
            for i in range (0, maxMM):
                if self.GetMM(i):
                    LResult = LResult+LUStrUtils.AddChar(' ', str(i+1), 2)+','
                else:
                    LResult = LResult + 'xx' + ','
                #endif
            #endfor
        #endif
        if XXName.upper() == 'DW':
            for i in range (0, maxDW):
                if self.GetDW(i):
                    LResult = LResult+LUStrUtils.AddChar(' ', str(i+1), 2)+','
                else:
                    LResult = LResult + 'xx' + ','
                #endif
            #endfor
        #endif
        if XXName.upper() == 'HH':
            for i in range (0, maxHH+1):
                if self.GetHH(i):
                    LResult = LResult+LUStrUtils.AddChar(' ', str(i), 2)+','
                else:
                    LResult = LResult + 'xx' + ','
                #endif
            #endfor
        #endif
        if XXName.upper() == 'NN':
            for i in range (0, maxNN+1):
                if self.GetNN(i):
                    LResult = LResult+LUStrUtils.AddChar(' ', str(i), 2)+','
                else:
                    LResult = LResult + 'xx' + ','
                #endif
            #endfor
        #endif
        # удалить последний символ ',' в строке
        LResult = LResult [:-1]
        return LResult
    #endfunction

    def CreateList (self, Pattern: str, Lmin: int, Lmax: int):
    #var
    #   i, j, WCount, i11, i12: Integer;
    #   S1, S11, S12: string;
    #beginfunction
        self.__FList.clear()
        if Pattern == '*':
            for j in range (Lmin, Lmax+1):
                # self.__FList.append (str(j))
                self.__FList.append (j)
        else:
            WCount = LUStrUtils.WordCount(Pattern, DelimItem)
            for i in range (1, WCount+1):
                s1 = LUStrUtils.ExtractWord (i, Pattern, DelimItem)
                s11 = LUStrUtils.ExtractWord (1, s1, DelimItems)
                s12 = LUStrUtils.ExtractWord (2, s1, DelimItems)
                try:
                    i11 = int(s11)
                except:
                    i11 = -1
                #endtry
                if i11 > Lmax: i11 = Lmax #endif
                try:
                    i12 = int(s12)
                except:
                    i12 = i11
                #endtry
                if i12 > Lmax: i12 = Lmax #endif
                if i11 >= 0 and i12 >= 0:
                    # for j in range (i11, i12 + 1): self.__FList.append (str (j))
                    for j in range (i11, i12 + 1): self.__FList.append (j)
                #endif
            #endfor
        #endif
    #endfunction

    # def NewPatterns (self, Patterns: str):
    # #beginfunction
    #    self.__FDD = 0
    #    self.__FMM = 0
    #    self.__FDW = 0
    #    self.__FHH = 0
    #    self.__FNN1 = 0
    #    self.__FNN2 = 0
    #    self.AddPatterns (Patterns)
    # #endfunction

    def AddPatterns (self, Patterns: str):
    #beginfunction
        self.__FDD = 0
        self.__FMM = 0
        self.__FDW = 0
        self.__FHH = 0
        self.__FNN1 = 0
        self.__FNN2 = 0
        # NN
        self.CreateList (LUStrUtils.ExtractWord (1, Patterns, DelimPattern), minNN, maxNN)
        for item in self.__FList:
            # self.SetNN(self.__FList.index(item), True)
            self.SetNN(item, True)
        # HH
        self.CreateList (LUStrUtils.ExtractWord (2, Patterns, DelimPattern), minHH, maxHH)
        for item in self.__FList:
            # self.SetHH (self.__FList.index(item), True)
            self.SetHH (item, True)
        # DD
        self.CreateList (LUStrUtils.ExtractWord (3, Patterns, DelimPattern), minDD, maxDD)
        for item in self.__FList:
            # self.SetDD (self.__FList.index(item), True)
            self.SetDD (item-1, True)
        # MM
        self.CreateList (LUStrUtils.ExtractWord (4, Patterns, DelimPattern), minMM, maxMM)
        for item in self.__FList:
            # self.SetMM (self.__FList.index(item), True)
            self.SetMM (item-1, True)
        # DW
        self.CreateList (LUStrUtils.ExtractWord (5, Patterns, DelimPattern), minDW, maxDW)
        for item in self.__FList:
            # self.SetDW (self.__FList.index (item), True)
            self.SetDW (item-1, True)
    #endfunction
#endclass

# --------------------------------------------
# TSheduler
# --------------------------------------------
class TSheduler (object):
    """TSheduler"""
    luClassName = 'TSheduler'

    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, *args, **kwargs):
        """Constructor"""
    #beginfunction
        super ().__init__ (*args, **kwargs)
        self.args = args
        self.kwargs = kwargs
        #
        self.__FStopThread = False
        #
        self.__FShedulerEvents = list ()
        #
        self.__FNameEvents = list ()
        # 'Следующий сеанс: ', self.DTEvents
        self.__FDTEvents: datetime = 0
        #
        self.__FEnable: bool = False
        #
        self.__FOnSheduler: TNotifyFileEvent = None
        #
        self.__FOnShedulerEvent:TShedulerEventItem = None

    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__(self):
        """destructor"""
    #beginfunction
        # удалить объект
        del self.__FShedulerEvents
        del self.__FNameEvents
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format(LClassName)
        # LUConst.LULogger.log (LULog.DEBUGTEXT, s)
    #endfunction

    #--------------------------------------------------
    # @property OnSheduler
    #--------------------------------------------------
    # getter
    @property
    def OnSheduler(self) -> TNotifyFileEvent:
    #beginfunction
        return self.__FOnSheduler
    #endfunction
    @OnSheduler.setter
    def OnSheduler(self, Value):
    #beginfunction
        self.__FOnSheduler = Value
        self.__FFunction = Value
    #endfunction

    #--------------------------------------------------
    # @property Enable
    #--------------------------------------------------
    # getter
    @property
    def Enable(self) -> bool:
    #beginfunction
        return self.__FEnable
    #endfunction
    @Enable.setter
    def Enable(self, Value: bool):
    #beginfunction
        self.__FEnable: bool = Value
        if Value:
            self.CreateNextEvent (True)
        #endif
    #endfunction

    # #--------------------------------------------------
    # # @property ShedulerEvents
    # #--------------------------------------------------
    # # getter
    # @property
    # def ShedulerEvents(self):
    # #beginfunction
    #     return self.__FShedulerEvents
    # #endfunction
    # @ShedulerEvents.setter
    # def ShedulerEvents(self, Value: ()):
    # #beginfunction
    #     self.__FShedulerEvents: () = Value
    # #endfunction

    #--------------------------------------------------
    # @property OnShedulerEvent
    #--------------------------------------------------
    # getter
    @property
    def OnShedulerEvent(self) -> TShedulerEventItem:
    #beginfunction
        return self.__FOnShedulerEvent
    #endfunction
    @OnShedulerEvent.setter
    def OnShedulerEvent(self, Value:TShedulerEventItem):
    #beginfunction
        self.__FOnShedulerEvent = Value
    #endfunction

    #--------------------------------------------------
    # @property ShedulerEventItem
    #--------------------------------------------------
    def GetShedulerEventItem(self, ANameEvent: str) -> TShedulerEventItem:
    #beginfunction
        for item in self.__FShedulerEvents:
            if item.NameEvent == ANameEvent:
                return item
            #endif
        #endfor
        return None
    #endfunction

    # #--------------------------------------------------
    # # @property NameEvents
    # #--------------------------------------------------
    # # getter
    # @property
    # def NameEvents(self) -> list:
    # #beginfunction
    #     self.CreateNextEvent (True)
    #     return self.__FNameEvents
    # #endfunction

    #--------------------------------------------------
    # @property DTEvents
    #--------------------------------------------------
    # getter
    @property
    def DTEvents(self) -> datetime:
    #beginfunction
        return self.__FDTEvents
    #endfunction

    #--------------------------------------------------
    # __run_Function
    #--------------------------------------------------
    def run_Function(self):
        """__run_Function"""
    #beginfunction
        # s = '__run_Function...'
        # LULog.LoggerTOOLS_AddDebug (s)
        if self.__FEnable:
            LPresentNow = LUDateTime.Now ()
            LYear, LMonth, LDay = LUDateTime.DecodeDate (LPresentNow)
            LHour, LMin, LSec, LMSec = LUDateTime.DecodeTime (LPresentNow)
            LDTEventsNow = LUDateTime.EncodeDateTime (LYear, LMonth, LDay, LHour, LMin, 0, 0)
            # print (LDTEventsNow, self.DTEvents)
            if LDTEventsNow == self.DTEvents:
                self.__Second ()
                # 'Следующий сеанс: ', self.DTEvents
                self.CreateNextEvent (False)
            #endif
        #endif
    #endfunction

    def CreateNextEvent (self, APrint: bool):
        """__CreateNextEvent"""
    #beginfunction
        LPresent: datetime = LUDateTime.Now ()
        LYear, LMonth, LDay = LUDateTime.DecodeDate (LPresent)
        LHour, LMin, LSec, LMSec = LUDateTime.DecodeTime (LPresent)
        LMin = LMin + 1
        self.__FNameEvents.clear()
        # Year
        for LYear in range (LYear,LYear+1):
            # Month
            for LMonth in range (LMonth, maxMM):
                # MaxDDWork = DateUtils.DaysInMonth (Present)
                MaxDDWork = monthrange (LYear, LMonth)[1]
                # Day of Month
                for LDay in range (LDay, MaxDDWork+1):
                    LDayWeek = LUDateTime.DayOfWeek (LUDateTime.EncodeDate(LYear,LMonth,LDay))
                    LDayWeek = LDayWeek + 1
                    if LDayWeek == 0:
                        LDayWeek = 7
                    #endif
                    # Hour
                    for LHour in range (LHour, maxHH+1):
                        # Min
                        for LMin in range (LMin, maxNN+1):
                            # Check List Sheduler
                            for item in  self.__FShedulerEvents:
                                self.OnShedulerEvent:TShedulerEventItem = item
                                bDay = self.OnShedulerEvent.GetDD (LDay-1)
                                bDayWeek = self.OnShedulerEvent.GetDW (LDayWeek-1)
                                bMonth = self.OnShedulerEvent.GetMM (LMonth-1)
                                bHour = self.OnShedulerEvent.GetHH (LHour)
                                bMin = self.OnShedulerEvent.GetNN (LMin)
                                # print (bDay, bDayWeek, bMonth, bHour, bMin)
                                if bDay and bDayWeek and bMonth and bHour and bMin:
                                     self.__FNameEvents.append (self.OnShedulerEvent.NameEvent)
                                #endif
                            #endfor
                            if len(self.__FNameEvents) > 0:
                                break
                            #endif
                        #endfor
                        if len(self.__FNameEvents) == 0:
                            LMin = minNN
                        else:
                            break
                        #endif
                    #endfor
                    if len(self.__FNameEvents) == 0:
                        LHour = minHH
                    else:
                        break
                    #endif
                #endfor
                if len(self.__FNameEvents) == 0:
                    LDay = minDD
                else:
                    break
                #endif
            #endfor
            if len(self.__FNameEvents) == 0:
                LMonth = minMM
            else:
                break
            #endif
        #endfor
        if len(self.__FNameEvents) > 0:
            D = LUDateTime.EncodeDate (LYear, LMonth, LDay)
            T = LUDateTime.EncodeTime (LHour, LMin, 0, 0)
            self.__FDTEvents = datetime.datetime.combine (D, T)
            self.__FDTEvents = LUDateTime.EncodeDateTime (LYear, LMonth, LDay, LHour, LMin, 0, 0)
        else:
            self.__FDTEvents = LPresent
        #endif
        s = 'Следующий сеанс: ' + str (self.__FDTEvents)
        if APrint:
            LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        #endif
    #endfunction
    
    def __Second (self):
        """Second"""
    #beginfunction
        # s = 'Second...'
        # LULog.LoggerTOOLS_AddDebug (s)
        if self.__FEnable:
            LPresent: datetime = LUDateTime.Now()
            LHour, LMin, LSec, LMSec = LUDateTime.DecodeTime (LPresent)
            # Check List Sheduler
            LYear, LMonth, LDay = LUDateTime.DecodeDate (LPresent)
            LDayWeek = LUDateTime.DayOfWeek (LPresent)
            LDayWeek = LDayWeek + 1
            if LDayWeek == 0:
                LDayWeek = 7
            for item in self.__FShedulerEvents:
                LEvent: TShedulerEventItem = item
                bDay = LEvent.GetDD (LDay-1)
                bDayWeek = LEvent.GetDW (LDayWeek-1)
                bMonth = LEvent.GetMM (LMonth-1)
                bHour = LEvent.GetHH (LHour)
                bMin = LEvent.GetNN (LMin)
                if bDay and bDayWeek and bMonth and bHour and bMin:
                    self.OnShedulerEvent = item
                    if self.OnSheduler is not None:
                        self.OnSheduler (LEvent)
                    #endif
                #endif
            #endfor
        #endif
    #endfunction

    def __CreateShedulerEvent(self) -> TShedulerEventItem:
    #beginfunction
        LResult: TShedulerEventItem = TShedulerEventItem()
        self.__FShedulerEvents.append(LResult)
        return LResult
    #endfunction

    def __DeleteEvent (self, ANameEvent: str):
    #beginfunction
        for item in self.__FShedulerEvents:
            if item.NameEvent == ANameEvent:
                self.__FShedulerEvents.remove(item)
                break
            #endif
        #endfor
    #endfunction

    def AddEvent (self, ANameEvent:str, APatterns: str) -> TShedulerEventItem:
        """AddEvent"""
    #beginfunction
        LEvent:TShedulerEventItem = None
        for item in self.__FShedulerEvents:
            if item.NameEvent == ANameEvent:
                LEvent: TShedulerEventItem = item
                break
            #endif
        #endfor

        if LEvent is not None and LEvent.NameEvent == ANameEvent:
            LEvent.AddPatterns (APatterns)
        else:
            LEvent:TShedulerEventItem = self.__CreateShedulerEvent()
            LEvent.NameEvent = ANameEvent
            LEvent.AddPatterns (APatterns)
        #endif
        return LEvent
    #endfunction

    @staticmethod
    def PrintEvent (AEvent: TShedulerEventItem):
        """PrintEvent"""
    #beginfunction
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, AEvent.NameEvent)
        s = f'NN={AEvent.GetXXString ("NN")}'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        s = f'HH={AEvent.GetXXString ("HH")}'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        s = f'DD={AEvent.GetXXString ("DD")}'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        s = f'MM={AEvent.GetXXString ("MM")}'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        s = f'DW={AEvent.GetXXString ("DW")}'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
    #endfunction
#endclass

# --------------------------------------------
# TShedulerThread
# --------------------------------------------
class TShedulerThread (threading.Thread):
    """TShedulerThread"""
    luClassName = 'TShedulerThread'

    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, *args, **kwargs):
        """Constructor"""
    #beginfunction
        super ().__init__ (*args, **kwargs)
        self.args = args
        self.kwargs = kwargs
        #
        self.__FStopThread = False
        self.__FSheduler = TSheduler()
    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__(self):
        """destructor"""
    #beginfunction
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format(LClassName)
        # LUConst.LULogger.log (LULog.DEBUGTEXT, s)
    #endfunction

    #--------------------------------------------------
    # @property Sheduler
    #--------------------------------------------------
    # getter
    @property
    def Sheduler(self) -> TSheduler:
    #beginfunction
        return self.__FSheduler
    #endfunction


    #--------------------------------------------------
    # run
    #--------------------------------------------------
    def run(self):
        """run - Запуск потока TSheduler"""
    #beginfunction
        s = 'run - Запуск потока TSheduler...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        super ().run()

        if self.__FSheduler.DTEvents == 0:
            self.__FSheduler.CreateNextEvent (False)
        #endif
        while not self.__FStopThread:
            self.__FSheduler.run_Function ()
            continue
        #endwhile
    #endfunction

    #--------------------------------------------------
    # StartSheduler
    #--------------------------------------------------
    def StartSheduler(self):
        """StartSheduler"""
    #beginfunction
        s = 'StartSheduler...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        self.__FStopThread = False
        self.__FSheduler.Enable = True
        # self.__CreateNextEvent ()
        self.start ()
    #endfunction
    #--------------------------------------------------
    # StopSheduler
    #--------------------------------------------------
    def StopSheduler(self):
        """StopSheduler"""
    #beginfunction
        s = 'StopSheduler...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        self.__FStopThread = True
        self.__FSheduler.Enable = False
    #endfunction
#endclass

# --------------------------------------------
# TShedulerTimer
# --------------------------------------------
class TShedulerTimer (threading.Timer):
    """TSheduler"""
    luClassName = 'TShedulerTimer'

    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, AInterval, AFunction, *args, **kwargs):
        """Constructor"""
        #beginfunction
        super ().__init__ (AInterval, AFunction, *args, **kwargs)
        self.args = args
        self.kwargs = kwargs
        #
        self.__FSheduler = TSheduler()
        #
        self.__FStopThread = False
        self.__FInterval = AInterval
        self.__FFunction = AFunction
    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__ (self):
        """destructor"""
    #beginfunction
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format (LClassName)
        # LUConst.LULogger.log (LULog.DEBUGTEXT, s)
    #endfunction

    #--------------------------------------------------
    # @property Sheduler
    #--------------------------------------------------
    # getter
    @property
    def Sheduler(self) -> TSheduler:
    #beginfunction
        return self.__FSheduler
    #endfunction

    #--------------------------------------------------
    # run
    #--------------------------------------------------
    def run (self):
        """run - Запуск таймера TShedulerTimer"""
        #beginfunction
        s = 'run - Запуск таймера TShedulerTimer...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)

        if self.__FSheduler.DTEvents == 0:
            self.__FSheduler.CreateNextEvent (True)
        #endif
        while not self.__FStopThread:
            self.__FSheduler.run_Function ()
            continue
        #endwhile
    #endfunction

    #--------------------------------------------------
    # StartSheduler
    #--------------------------------------------------
    def StartSheduler (self):
        """StartSheduler"""
    #beginfunction
        s = 'StartSheduler...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        self.__FStopThread = False
        self.__FSheduler.Enable = True
        # self.__CreateNextEvent (True)
        self.start ()
    #endfunction
    #--------------------------------------------------
    # StopSheduler
    #--------------------------------------------------
    def StopSheduler (self):
        """StopSheduler"""
    #beginfunction
        s = 'StopSheduler...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)
        self.__FStopThread = True
        self.__FSheduler.Enable = False
    #endfunction
#endclass

#---------------------------------------------------------
# main
#---------------------------------------------------------
def main ():
#beginfunction
    ...
#endfunction

#---------------------------------------------------------
#
#---------------------------------------------------------
#beginmodule
if __name__ == "__main__":
    main()
#endif

#endmodule




"""LUParserREG.py"""
# -*- coding: UTF-8 -*-
__annotations__ = """
 =======================================================
 Copyright (c) 2023
 Author:
     Lisitsin Y.R.
 Project:
     LU_PY
     Python (LU)
 Module:
     LUParserREG.py

 =======================================================
"""

#------------------------------------------
# БИБЛИОТЕКИ python
#------------------------------------------
import enum
import platform

if platform.system() == 'Windows':
    import winreg
#endif

#------------------------------------------
# БИБЛИОТЕКИ сторонние
#------------------------------------------

#------------------------------------------
# БИБЛИОТЕКА LU
#------------------------------------------

#----------------------------------------------------------
# HKEY_* Constants
#----------------------------------------------------------
"""
winreg.HKEY_CLASSES_ROOT
Registry entries subordinate to this key define types (or classes) of documents and the properties associated with those types. Shell and COM applications use the information stored under this key.
---------------------
winreg.HKEY_CURRENT_USER
Registry entries subordinate to this key define the preferences of the current user. These preferences include the settings of environment variables, data about program groups, colors, printers, network connections, and application preferences.
---------------------
winreg.HKEY_LOCAL_MACHINE
Registry entries subordinate to this key define the physical state of the computer, including data about the bus type, system memory, and installed hardware and software.
---------------------
winreg.HKEY_USERS
Registry entries subordinate to this key define the default user configuration for new users on the local computer and the user configuration for the current user.
---------------------
winreg.HKEY_PERFORMANCE_DATA
Registry entries subordinate to this key allow you to access performance data. The data is not actually stored in the registry; the registry functions cause the system to collect the data from its source.
---------------------
winreg.HKEY_CURRENT_CONFIG
Contains information about the current hardware profile of the local computer system.
---------------------
winreg.HKEY_DYN_DATA
This key is not used in versions of Windows after 98.
---------------------
"""
@enum.unique
class THKEYConst(enum.Enum):
    """THKEYConst"""
    cHKCR = winreg.HKEY_CLASSES_ROOT
    cHKCU = winreg.HKEY_CURRENT_USER
    cHKLM = winreg.HKEY_LOCAL_MACHINE
    cHKU = winreg.HKEY_USERS
    cHKPD = winreg.HKEY_PERFORMANCE_DATA
    cHKCC = winreg.HKEY_CURRENT_CONFIG
    cHKDD = winreg.HKEY_DYN_DATA
#endclass

#---------------------------------------------------------
# Access Rights
#---------------------------------------------------------
"""
winreg.KEY_ALL_ACCESS
Combines the STANDARD_RIGHTS_REQUIRED, KEY_QUERY_VALUE, KEY_SET_VALUE, KEY_CREATE_SUB_KEY, KEY_ENUMERATE_SUB_KEYS, KEY_NOTIFY, and KEY_CREATE_LINK access rights.
---------------------
winreg.KEY_WRITE
Combines the STANDARD_RIGHTS_WRITE, KEY_SET_VALUE, and KEY_CREATE_SUB_KEY access rights.
---------------------
winreg.KEY_READ
Combines the STANDARD_RIGHTS_READ, KEY_QUERY_VALUE, KEY_ENUMERATE_SUB_KEYS, and KEY_NOTIFY values.
---------------------
winreg.KEY_EXECUTE
Equivalent to KEY_READ.
---------------------
winreg.KEY_QUERY_VALUE
Required to query the values of a registry key.
---------------------
winreg.KEY_SET_VALUE
Required to create, delete, or set a registry value.
---------------------
winreg.KEY_CREATE_SUB_KEY
Required to create a subkey of a registry key.
---------------------
winreg.KEY_ENUMERATE_SUB_KEYS
Required to enumerate the subkeys of a registry key.
---------------------
winreg.KEY_NOTIFY
Required to request change notifications for a registry key or for subkeys of a registry key.
---------------------
winreg.KEY_CREATE_LINK
Reserved for system use.
---------------------
"""
# @enum.unique
class TKEYAccess(enum.Enum):
    """TKEYAccess"""
    kaALL_ACCESS = winreg.KEY_ALL_ACCESS
    kaWRITE = winreg.KEY_WRITE
    kaREAD = winreg.KEY_READ
    kaEXECUTE = winreg.KEY_EXECUTE
    kaQUERY_VALUE = winreg.KEY_QUERY_VALUE
    kaSET_VALUE = winreg.KEY_SET_VALUE
    kaCREATE_SUB_KEY = winreg.KEY_CREATE_SUB_KEY
    kaENUMERATE_SUB_KEYS = winreg.KEY_ENUMERATE_SUB_KEYS
    kaKEY_NOTIFY = winreg.KEY_NOTIFY
    kaKEY_CREATE_LINK = winreg.KEY_CREATE_LINK
#endclass

#---------------------------------------------------------
# Value Types¶
# ---------------------------------------------------------
"""
winreg.REG_BINARY
Binary data in any form.
---------------------
winreg.REG_DWORD
32-bit number.
---------------------
winreg.REG_DWORD_LITTLE_ENDIAN
A 32-bit number in little-endian format. Equivalent to REG_DWORD.
---------------------
winreg.REG_DWORD_BIG_ENDIAN
A 32-bit number in big-endian format.
---------------------
winreg.REG_EXPAND_SZ
Null-terminated string containing references to environment variables (%PATH%).
---------------------
winreg.REG_LINK
A Unicode symbolic link.
---------------------
winreg.REG_MULTI_SZ
A sequence of null-terminated strings, terminated by two null characters. (Python handles this termination automatically.)
---------------------
winreg.REG_NONE
No defined value type.
---------------------
"""
# @enum.unique
class TValueTypes(enum.Enum):
    """TValueTypes"""
    vtBINARY = winreg.REG_BINARY
    vtDWORD = winreg.REG_DWORD
    vtDWORD_LITTLE_ENDIAN = winreg.REG_DWORD_LITTLE_ENDIAN
    vtDWORD_BIG_ENDIAN = winreg.REG_DWORD_BIG_ENDIAN
    vtEXPAND_SZ = winreg.REG_EXPAND_SZ
    vtLINK = winreg.REG_LINK
    vtMULTI_SZ = winreg.REG_MULTI_SZ
    vtNONE = winreg.REG_NONE
#endclass

#-------------------------------------------------------------------------------
# General Reestr Keys
#-------------------------------------------------------------------------------
RootKeyHKLM = winreg.HKEY_LOCAL_MACHINE
RootKeyHKCU = winreg.HKEY_CURRENT_USER

cHKLMSCCS    = r'HKEY_LOCAL_MACHINE\SYSTEM\CurrentControlSet'
cHKLMSMWCV   = r'HKEY_LOCAL_MACHINE\Software\Microsoft\Windows\CurrentVersion'
cHKLMSMWNTCV = r'HKEY_LOCAL_MACHINE\Software\Microsoft\Windows NT\CurrentVersion'
cHKCUSMWCV   = r'HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion'
cHKCUSMWNTCV = r'HKEY_CURRENT_USER\Software\Microsoft\Windows NT\CurrentVersion'

class TREGParser (object):
    """TREGParser"""
    luClassName = 'TREGParser'
    __annotations__ =\
    """
    TREGParser - Работа с реестром windows
    
    aReg = _winreg.ConnectRegistry(None,_winreg.HKEY_LOCAL_MACHINE)
    
    """
    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, **kwargs):
        """ Constructor """
    #beginfunction
        super ().__init__ (**kwargs)
        self.__Fhkey = 0
        ...
    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__ (self):
        """ destructor """
    #beginfunction
        LClassName = self.__class__.__name__
        s = '{} уничтожен'.format (LClassName)
        # LULog.LoggerTOOLS_AddLevel (LULog.DEBUGTEXT, s)
        #print (s)
    #endfunction

    #--------------------------------------------------
    # @property hkey
    #--------------------------------------------------
    # getter
    @property
    def hkey(self):
    #beginfunction
        return self.__Fhkey
    #endfunction
    # setter
    @hkey.setter
    def hkey (self, AValue: int):
    #beginfunction
        self.__Fhkey = AValue
    #endfunction

    def CreateKeyReg (self, AHKEY: THKEYConst, ASection: str) -> bool:
        """CreateKeyReg"""
    #beginfunction
        # winreg.CreateKey (key, sub_key)
        #   Создает или открывает указанный ключ, возвращая объект дескриптора
        # winreg.CreateKeyEx(key, sub_key, reserved=0, access=KEY_WRITE)
        #   Создает или открывает указанный ключ, возвращая объект дескриптора
        self.hkey = None
        try:
            self.hkey = winreg.CreateKeyEx (AHKEY.value, ASection, 0, TKEYAccess.kaALL_ACCESS.value)
            self.CloseKeyReg (self.hkey)
        except FileNotFoundError as ERROR:
            ...
        finally:
            # return not self.hkey is None
            pass

        return not self.hkey is None
    #endfunction

    @staticmethod
    def DeleteKeyReg (AHKEY: THKEYConst, ASection: str):
        """DeleteKeyReg"""
    #beginfunction
        # winreg.DeleteKey (key, sub_key)
        #   Удаляет указанный ключ.
        # winreg.DeleteKeyEx (key, sub_key, access = KEY_WOW64_64KEY, reserved = 0)
        #   Удаляет указанный ключ.
        LResult = False
        try:
            winreg.DeleteKey (AHKEY.value, ASection)
            LResult = True
        except FileNotFoundError as ERROR:
            ...
        finally:
            # return LResult
            pass
        return LResult
    #endfunction

    def OpenKeyReg (self, AHKEY: THKEYConst, ASection: str):
        """OpenKeyReg"""
    #beginfunction
        #+winreg.OpenKey (key, sub_key, reserved = 0, access = KEY_READ)
        #   Открывает указанный ключ, возвращая объект дескриптора .
        #+winreg.OpenKeyEx (key, sub_key, reserved = 0, access = KEY_READ)
        #   Открывает указанный ключ, возвращая объект дескриптора .
        self.hkey = None
        try:
            self.hkey = winreg.OpenKeyEx (AHKEY.value, ASection, 0, TKEYAccess.kaALL_ACCESS.value)
        except FileNotFoundError as ERROR:
            ...
        finally:
            # return self.hkey
            pass
        return self.hkey
    #endfunction

    @staticmethod
    def CloseKeyReg (Ahkey):
        """CloseKeyReg"""
    #beginfunction
        # winreg.CloseKey(hkey)
        # Закрывает ранее открытый раздел реестра. HKEY аргумент указывает, ранее открытый ключ.
        # Если hkey не закрыт с помощью этого метода (или через hkey.Close() ), он закрывается,
        # когда объект hkey уничтожается Python.
        LResult = False
        if Ahkey:
            winreg.CloseKey (Ahkey)
            LResult = True
        return LResult
    #endfunction

    def EnumKeyReg (self, AHKEY: THKEYConst, ASection: str) -> []:
        """EnumKeyReg"""
    #beginfunction
        # winreg.EnumKey (key, index)
        #   Нумеровывает подклавиши открытого ключа реестра,возвращая строку.
        LInfo = self.QueryInfoKeyReg (AHKEY, ASection)
        LList = []
        if len(LInfo) > 0:
            self.hkey = self.OpenKeyReg (AHKEY, ASection)
            # LWork = EnumKey (self.hkey, 0)
            for i in range (LInfo[0],LInfo[1]):
                try:
                    LWork = winreg.EnumKey (self.hkey, i)
                except OSError as ERROR:
                    LWork = ERROR.strerror
                    #LUErrors.LUFileError_FileNotExist as ERROR
                    ...
                finally:
                    ...
                LList.append(LWork)
            self.CloseKeyReg (self.hkey)
        return LList
    #endfunction

    def EnumValueReg (self, AHKEY: THKEYConst, ASection: str) -> []:
        """EnumValueReg"""
    #beginfunction
        #+winreg.EnumValue (key, index)
        #   Перечисляет значения открытого ключа реестра,возвращая кортеж.
        LInfo = self.QueryInfoKeyReg (AHKEY, ASection)
        LList = []
        if len(LInfo) > 0:
            self.hkey = self.OpenKeyReg (AHKEY, ASection)
            for i in range (LInfo[0],LInfo[1]):
                LList.append(winreg.EnumValue (self.hkey, i))
                # LKeyName, LValue, LFormat = EnumValue (self.hkey, i)
            self.CloseKeyReg (self.hkey)
        return LList
    #endfunction

    def SaveKeyReg (self, AHKEY: THKEYConst, ASection: str, AFileName: str):
        """SaveKeyReg"""
    #beginfunction
        # winreg.SaveKey (key, file_name)
        #   Сохраняет указанный ключ и все его подклавиши в указанный файл.
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            winreg.SaveKey (self.hkey, AFileName)
            self.CloseKeyReg(self.hkey)
    #endfunction

    def LoadKeyReg (self, AHKEY: THKEYConst, ASection: str, AFileName: str):
        """LoadKeyReg"""
    #beginfunction
        # winreg.LoadKey(key, sub_key, file_name)
        #   Создает под-ключ под указанным ключом и сохраняет регистрационную информацию из указанного файла в этот под-ключ.
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            winreg.LoadKey (self.hkey, ASection, AFileName)
            self.CloseKeyReg(self.hkey)
    #endfunction

    def QueryValueReg (self, AHKEY: THKEYConst, ASection: str, AOption: str) -> ():
        """QueryValueReg"""
    #beginfunction
        #+winreg.QueryValue (key, sub_key)
        #   Возвращает безымянное значение для ключа,в виде строки.
        #+winreg.QueryValueEx (key, value_name)
        #   Результат-кортеж из 2 пунктов:
        #   0 - Значение элемента реестра.
        #   1 - Целое число, указывающее тип реестра для этого значения
        LResult = ''
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            if len(AOption) > 0:
                LResult = winreg.QueryValueEx (self.hkey, AOption)
            else:
                LResult = winreg.QueryValue (self.hkey, None)
            self.CloseKeyReg(self.hkey)
        return LResult
    #endfunction

    def QueryInfoKeyReg (self, AHKEY: THKEYConst, ASection: str) -> ():
        """QueryInfoKeyReg"""
    #beginfunction
        #+winreg.QueryInfoKey (key)
        #   Результат-кортеж из 3 пунктов:
        LResult = ()
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            LResult = winreg.QueryInfoKey (self.hkey)
            self.CloseKeyReg(self.hkey)
        return LResult
    #endfunction

    def DeleteValueReg (self, AHKEY: THKEYConst, ASection: str, AOption: str) -> bool:
        """SetValueReg"""
    #beginfunction
        # winreg.DeleteValue (key, value)
        #   Удаляет именованное значение из ключа реестра.
        LResult = False
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            winreg.DeleteValue (self.hkey, AOption)
            self.CloseKeyReg(self.hkey)
            LResult = True
        return LResult
    #endfunction

    def SetValueReg (self, AHKEY: THKEYConst, ASection: str, AOption: str, AFormat: TValueTypes, Value: str):
        """SetValueReg"""
    #beginfunction
        # winreg.SetValue (key, sub_key, type, value)
        #   Сопоставляет стоимость с указанным ключом.
        # winreg.SetValueEx (key, value_name, reserved, type, value)¶
        #   Хранит данные в поле значений открытого ключа реестра.
        LResult = False
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            winreg.SetValueEx (self.hkey, AOption, 0, AFormat.value, Value)
            self.CloseKeyReg(self.hkey)
            LResult = True
        return LResult
    #endfunction

    def GetKeyReg (self, AHKEY: THKEYConst, ASection: str, AOption: str):
        """GetKeyReg"""
    #beginfunction
        return self.QueryValueReg (AHKEY, ASection, AOption)
    #endfunction

    def GetOptionsReg (self, AHKEY: THKEYConst, ASection: str) -> ():
        """QueryInfoKeyReg"""
    #beginfunction
        #+winreg.QueryInfoKey (key)
        #   Результат-кортеж из 3 пунктов:
        LListKeyValue = self.EnumValueReg (AHKEY, ASection)
        LList = []
        for key in LListKeyValue:
            LList.append(key[0])
        return LList
    #endfunction

    def IsSection (self, AHKEY: THKEYConst, ASection: str) -> bool:
    #beginfunction
        LResult = False
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            LResult =  True
        return LResult
    #endfunction

    def IsOption (self, AHKEY: THKEYConst, ASection: str, AOption: str) -> bool:
    #beginfunction
        LResult = False
        self.hkey = self.OpenKeyReg (AHKEY, ASection)
        if not self.hkey is None:
            LList = self.GetOptionsReg (AHKEY, ASection)
            if len (LList) > 0 and AOption in LList:
                LResult = True
            #endif
        #endif
        return LResult
    #endfunction
#endclass

def SaveRegToFile_regedit (AFileName: str, AHKEY: THKEYConst, ASection: str):
    """SaveRegToFile"""
#beginfunction
    # LWorkDir = LUFile.ExtractFileDir (AFileName)
    LProgramName = 'regedit.exe'
    LParamStr = ''
    match AHKEY:
        case THKEYConst.cHKLM:
            s = 'HKEY_LOCAL_MACHINE'
            LParamStr = '/ea'+' '+AFileName+' '+s+'\\'+ASection
        case THKEYConst.cHKCU:
            s = 'HKEY_CURRENT_USER'
            LParamStr = '/ea'+' '+AFileName+' "'+s+'\\'+ASection+'"'
    #endmatch
    if len (LParamStr) > 0:
        #print (LParamStr)
        # Lregedit = subprocess.Popen ('C:\\Windows\\System32\\regedit.exe', LParamStr)
        # Lregedit = subprocess.Popen ('regedit.exe', LParamStr)
        # Lregedit = subprocess.Popen ('regedit.exe')
        # os.system (command)
        # os.startfile ('regedit.exe', LParamStr)
        # os.startfile ('regedit.exe')
        ...
    #endif
#endfunction

#---------------------------------------------------------
# CreateTREGParser
#---------------------------------------------------------
def CreateTREGParser () -> TREGParser:
    """CreateTREGParser"""
#beginfunction
    return TREGParser ()
#endfunction

GREGParser = CreateTREGParser ()

#---------------------------------------------------------
# main
#---------------------------------------------------------
def main ():
#beginfunction
    # print (tuple(THKEYConst))
    # print (tuple(TKEYAccess))
    # print (tuple(TValueTypes))
    ...
#endfunction

#---------------------------------------------------------
#
#---------------------------------------------------------
#beginmodule
if __name__ == '__main__':
    main()
#endif

#endmodule



"""LUsys.py"""
# -*- coding: UTF-8 -*-
__annotations__ = """
 =======================================================
 Copyright (c) 2023-2025
 Author:
     Lisitsin Y.R.
 Project:
     LU_PY
     Python (LU)
 Module:
     LUDecotators.py

 =======================================================
"""

#------------------------------------------
# БИБЛИОТЕКИ python
#------------------------------------------
import time
from functools import wraps
import smtplib
import traceback
from email.mime.text import MIMEText

#------------------------------------------
# БИБЛИОТЕКИ сторонние
#------------------------------------------

#------------------------------------------
# БИБЛИОТЕКИ LU
#------------------------------------------
import lyrpy.LULog as LULog

#---------------------------------------------------------------
# TIMING
#---------------------------------------------------------------
def TIMING(func):
#beginfunction

    def wrapper(*args, **kwargs):
    #beginfunction
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        s = f"Функция {func.__name__} работала {end_time - start_time} секунд..."
        LULog.LoggerAdd (LULog.LoggerTOOLS, LULog.DEBUGTEXT, s)
        return result
    #endfunction

    return wrapper
#endfunction

#---------------------------------------------------------------
# retry
#---------------------------------------------------------------
"""
1. Декоратор retry
В проектах по обработке данных и разработке программного обеспечения очень много случаев, когда мы зависим от внешних систем. Не всё всегда находятся под нашим контролем.
Иногда происходят неожиданное событие, во время которых нам бы хотелось, чтобы внешняя система сама исправляла возникнувшие ошибки и перезапускалась.
Я предпочитаю реализовывать эту логику с помощью декоратора retry, который позволяет повторно выполнять программу через N-ное количество времени.
"""
def retry(max_tries=3, delay_seconds=1):
    def decorator_retry(func):
        @wraps(func)
        def wrapper_retry(*args, **kwargs):
            tries = 0
            while tries < max_tries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    tries += 1
                    if tries == max_tries:
                        raise e
                    time.sleep(delay_seconds)
        return wrapper_retry
    return decorator_retry
@retry(max_tries=5, delay_seconds=2)
def call_dummy_api():
    response = None
    # response = requests.get("https://jsonplaceholder.typicode.com/todos/1")
    return response
#endfunction

#---------------------------------------------------------------
# memoize
#---------------------------------------------------------------
"""
2. Результаты функции кэширования
Некоторые части нашего кода редко меняют своё поведение. Тем не менее, если такое всё-таки произойдёт, это может отнять большую часть наших вычислительных мощностей. В таких ситуациях мы можем использовать декоратор для кэширования вызовов функций
Функция будет запущена только один раз, если входные данные совпадают. При каждом последующем запуске результаты будут извлекаться из кэша. Следовательно, нам не нужно будет постоянно выполнять дорогостоящие вычисления.
"""
def memoize(func):
    cache = {}
    def wrapper(*args):
        if args in cache:
            return cache[args]
        else:
            result = func(*args)
            cache[args] = result
            return result
    return wrapper
#endfunction

#---------------------------------------------------------------
# email_on_failure
#---------------------------------------------------------------
"""
5. Декоратор Notification
Наконец, очень полезным декоратором в производственных системах является декоратор Notification.
Ещё раз, даже при нескольких повторных попытках хорошо протестированная кодовая база может потерпеть неудачу. И когда это произойдет, нам нужно сообщить кому-нибудь об этом, чтобы принять быстрые меры.
Это не ново, если вы когда-либо создавали конвейер данных и надеялись, что он всегда будет работать без перебоев.
Следующий декоратор отправляет электронное письмо всякий раз, когда выполнение внутренней функции завершается неудачей. В вашем случае это не обязательно должно быть уведомление по электронной почте. Вы можете настроить его для отправки уведомлений Teams / slack:
"""
def email_on_failure (sender_email, password, recipient_email):
    def decorator (func):
        def wrapper (*args, **kwargs):
            try:
                return func (*args, **kwargs)
            except Exception as e:
                # format the error message and traceback
                err_msg = f"Error: {str (e)}\n\nTraceback:\n{traceback.format_exc ()}"

                # create the email message
                message = MIMEText (err_msg)
                message ['Subject'] = f"{func.__name__} failed"
                message ['From'] = sender_email
                message ['To'] = recipient_email

                # send the email
                with smtplib.SMTP_SSL ('smtp.gmail.com', 465) as smtp:
                    smtp.login (sender_email, password)
                    smtp.sendmail (sender_email, recipient_email, message.as_string ())

                # re-raise the exception
                raise

        return wrapper
    return decorator
#endfunction

#---------------------------------------------------------------
# my_function
#---------------------------------------------------------------
@email_on_failure (sender_email = 'your_email@gmail.com', password = 'your_password',
                   recipient_email = 'recipient_email@gmail.com')
def my_function ():
    # code that might fail
    ...
#endfunction

#---------------------------------------------------------------
# timeit
#---------------------------------------------------------------
"""
To overcome this, created the @timeit decorator which allows you to measure the execution time of the method/function by just adding the @timeit decorator on the method.
@timeit decorator:
"""
def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print ('%r  %2.2f ms' % (method.__name__, (te - ts) * 1000))
        return result
    return timed
#endfunction

#---------------------------------------------------------------
# get_all_employee_details
#---------------------------------------------------------------
# Adding decorator to the method
@timeit
def get_all_employee_details(**kwargs):
    print ('employee details')
# The code will look like this after removing the redundant code.
# logtime_data = {}
# employees = Employee.get_all_employee_details(log_time=logtime_data)
# Hurray!! All that messy code is gone and now it looks simple and clear.
# log_time and log_name are optional. Make use of them accordingly when needed.

#---------------------------------------------------------------
# timeit
#---------------------------------------------------------------
def timeit(func):
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')
        return result
    return timeit_wrapper
#endfunction

@timeit
def calculate_something(num):
    """
    Simple function that returns sum of all numbers up to the square of num.
    """
    total = sum((x for x in range(0, num**2)))
    return total
#endfunction

class Calculator:
    @timeit
    def calculate_something(self, num):
        """
        an example function that returns sum of all numbers up to the square of num
        """
        total = sum((x for x in range(0, num**2)))
        return total

    def __repr__(self):
        return f'calc_object:{id(self)}'

#---------------------------------------------------------------
#
#---------------------------------------------------------------
"""
01.Свойства @property
Декоратор @property облегчает создание свойств в классах Python. Свойства выглядят как обычные атрибуты (поля) класса, но при их чтении вызывается геттер (getter), при записи – сеттер (setter), а при удалении – делитер (deleter). Геттер и делитер опциональны.

02.Статические и классовые методы
Методы могут быть не только у экземпляра класса, но и у самого класса, которые вызываются без какого-то экземпляра (без self). Декораторы @staticmethod и @classmethod как раз делают метод таким (статическим или классовым). Эти декораторы встроены и видны без import.

Статический метод – это способ поместить функцию в класс, если она логически относится к этому классу. Статический метод ничего не знает о классе, из которого его вызвали.

class Foo:
    @staticmethod
    def help():
        print('help for Foo class')
Foo.help()

Классовый метод напротив знает, из какого класса его вызывают. Он принимает неявный первый аргумент (обычно его зовут cls), который содержит вызывающий класс. Классовые методы прекрасно подходят, когда нужно учесть иерархию наследования. Пример: метод group создает список из нескольких людей. Причем для Person – список Person, а для Worker – список Worker. Со @staticmethod такое бы не вышло:

class Person:
    @classmethod
    def group(cls, n):
        # cls именно тот класс, который вызвал
        return [cls() for _ in range(n)]
    def __repr__(self):
        return 'Person'
class Worker(Person):
    def __repr__(self):
        return 'Worker'
print(Person.group(3))
# [Person, Person, Person]
print(Worker.group(2))
# [Worker, Worker]

03.@contextmanager
Этот декоратор позволяет получить из генератора – контекст менеджер. Находится в стандартном модуле contextlib. Пример открытие файла.

from contextlib import contextmanager
@contextmanager
def my_open(name, mode='r'):
    # тут код для получения ресурса
    f = open(name, mode)
    print('Файл открыт:', name)
    try:
        yield f
    finally:
        # Code to release resource, e.g.:
        f.close()
        print('Файл закрыт:', name)
# использование
with my_open('1.txt', 'w') as f:
    f.write('Hello')
    f.fooooo()  # <- error
# Файл открыт: 1.txt
# Traceback (most recent call last):
# Файл закрыт: 1.txt

В этом генераторе есть единственный yield – он возвращает как раз нужный ресурс. Все, что до него – код захвата ресурса (будет выполнен в методе __enter__), например, открытие файла. Мы оборачиваем yield в try/finally, чтобы если даже в блоке кода with произойдет ошибка, то исключение выбросится из yield, но код закрытия файла в блоке finally будет выполнен в любом случае. Код закрытия выполняется в методе __exit__ менеджера контекста.

Асинхронная версия этого декоратора – @asynccontextmanager. Пример:

from contextlib import asynccontextmanager
@asynccontextmanager
async def get_connection():
    conn = await acquire_db_connection()
    try:
        yield conn
    finally:
        await release_db_connection(conn)
# использование
async def get_all_users():
    async with get_connection() as conn:
        return conn.query('SELECT ...')
        
@functools.wraps
Декоратор @functools.wraps полезен при разработке других декораторов. Передает имя, документацию и прочую мета-информацию из декорируемой функции к ее обертке. Подробнее в статье про декораторы.

@atexit.register
Декоратор @atexit.register регистрирует функцию для вызова ее при завершении работы процесса Python.

import atexit
@atexit.register
def goodbye():
    print("You are now leaving the Python sector.")
Измерение времени @timeit
Переходим к самописным декораторам.

Этот декоратор измеряет время выполнения функции, которую декорирует.

import time
from functools import wraps
def timeit(method):
    @wraps(method)
    def timed(*args, **kw):
        ts = time.monotonic()
        result = method(*args, **kw)
        te = time.monotonic()
        ms = (te - ts) * 1000
        all_args = ', '.join(tuple(f'{a!r}' for a in args)
                             + tuple(f'{k}={v!r}' for k, v in kw.items()))
        print(f'{method.__name__}({all_args}): {ms:2.2f} ms')
        return result
    return timed
# использование:
@timeit
def slow_func(x, y, sleep):
    time.sleep(sleep)
    return x + y
slow_func(10, 20, sleep=2)
# печатает: slow_func(10, 20, sleep=2): 2004.65 ms

Как видите, нам не нужно вмешиваться в код функции, не нужно каждый раз писать измеритель времени, декоратор отлично экономит нашу работу: надо измерить время – дописали @timeit и видим все, как на ладони.        

Повторитель
Повторяет вызов функции n раз, возвращает последний результат.

from functools import wraps
def repeat(_func=None, *, num_times=2):
    def decorator_repeat(func):
        @wraps(func)
        def wrapper_repeat(*args, **kwargs):
            value = None
            for _ in range(num_times):
                value = func(*args, **kwargs)
            return value
        return wrapper_repeat
    if _func is None:
        return decorator_repeat
    else:
        return decorator_repeat(_func)
@repeat(num_times=5)
def foo():
    print('текст')
"""

"""    
Замедлитель
Замедляет исполнение функции на нужное число секунд. Бывает полезно для отладки.

from functools import wraps
import time
def slow_down(seconds=1):
    def _slow_down(func):
        # Sleep 1 second before calling the function
        @wraps(func)
        def wrapper_slow_down(*args, **kwargs):
            time.sleep(seconds)
            return func(*args, **kwargs)
        return wrapper_slow_down
    return _slow_down
@slow_down(seconds=0.5)
def foo():
    print('foo')
def bar():
    foo()  # каждый foo по полсекунды
    foo()

Помощник для отладки
Этот декоратор будет логгировать все вызовы функции и печатать ее аргументы и возвращаемое значение.

from functools import wraps
# Печатает сигнатуру вызова и возвращаемое значение
import functools
def debug(func):
    @wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        print(f"Calling {func.__name__}({signature})")
        value = func(*args, **kwargs)
        print(f"{func.__name__!r} returned {value!r}")
        return value
    return wrapper_debug
@debug
def testee(x, y):
    print(x + y)
"""


#------------------------------------------
def main ():
#beginfunction
    print('main LUDecotators.py ...')
#endfunction

#------------------------------------------
#
#------------------------------------------
#beginmodule
if __name__ == "__main__":
    main()
#endif

#endmodule



"""LUQThread.py"""
# -*- coding: UTF-8 -*-
__annotations__ = """
 =======================================================
 Copyright (c) 2023
 Author:
     Lisitsin Y.R.
 Project:
     LU_PY
     Python (LU)
 Module:
     LUQThread.py

 =======================================================
"""

import logging

#------------------------------------------
# БИБЛИОТЕКИ python
#------------------------------------------
import psutil

#------------------------------------------
# БИБЛИОТЕКИ сторонние
#------------------------------------------
import PySide6.QtCore as QtCore
from PySide6.QtCore import QObject, QThread, Signal, Slot

#------------------------------------------
# БИБЛИОТЕКА LU
#------------------------------------------
import lyrpy.LULog as LULog

# Signals must inherit QObject
class MySignals(QObject):
    signal_str = Signal(str)
    signal_int = Signal(int)
#endclass

# Create the Worker Thread
class TQThread (QThread):
    """TQThread"""
    luClassName = 'TQThread'

    #--------------------------------------------------
    # constructor
    #--------------------------------------------------
    def __init__ (self, parent = None):
    #beginfunction
        QThread.__init__ (self, parent = parent)
        # super ().__init__ (*args, **kwargs)
        # self.args = args
        # self.kwargs = kwargs

        self.__FStopThread = False

        # Instantiate signals and connect signals to the slots
        self.signals = MySignals ()
        self.signals.signal_str.connect (parent.update_str_field)
        self.signals.signal_int.connect (parent.update_int_field)

    #endfunction

    #--------------------------------------------------
    # destructor
    #--------------------------------------------------
    def __del__ (self):
        """destructor"""
    #beginfunction
        LClassName = self.__class__.__name__
        # s = '{} уничтожен'.format (LClassName)
        # LULog.LoggerTOOLS_AddLevel (LULog.DEBUGTEXT, s)
        #print (s)
    #endfunction

    #--------------------------------------------------
    # @property QThread
    #--------------------------------------------------
    # getter
    @property
    def QThread(self) -> QThread:
    #beginfunction
        return self
    #endfunction

    #--------------------------------------------------
    # run
    #--------------------------------------------------
    def run(self):
        """Запуск потока"""
    #beginfunction
        super ().run()
        s = 'run - Запуск потока...'
        LULog.LoggerAdd (LULog.LoggerTOOLS, logging.DEBUG, s)

        # Do something on the worker thread
        a = 1 + 1
        # Emit signals whenever you want
        self.signals.signal_int.emit (a)
        self.signals.signal_str.emit ("This text comes to Main thread from our Worker thread.")

        while not self.__FStopThread:
            # s = 'Выполнение потока...'
            # LULog.LoggerTOOLS_AddDebug (s)
            Lval = psutil.cpu_percent ()
            self.emit(QtCore.SIGNAL('CPU_VALUE'), Lval)
            continue
        #endwhile

        # # Do something on the worker thread
        # a = 1 + 1
        # # Emit signals whenever you want
        # self.signals.signal_int.emit (a)
        # self.signals.signal_str.emit ("This text comes to Main thread from our Worker thread.")
        # while 1:
        #     Lval = psutil.cpu_percent ()
        #     # self.emit(QtCore.SIGNAL('CPU_VALUE'), Lval)
        #     ...
        # #endwhile

    #endfunction
#endclass

#---------------------------------------------------------
# main
#---------------------------------------------------------
def main ():
#beginfunction
    ...
#endfunction

#---------------------------------------------------------
#
#---------------------------------------------------------
#beginmodule
if __name__ == "__main__":
    main()
#endif

#endmodule
